"""
`BeamProfiler` is a Python package for laser beam analysis and
characterization according to ISO 13694, ISO 11145, and other non-ISO
definitions commonly used in the industry.
"""


__author__ = """Wagno Alves Braganca Junior"""
__email__ = 'wagnojunior@gmail.com'
__version__ = '0.1.4'


# # =============================================================================
# # Imports
# # =============================================================================
# from src.beamprofiler import analysis
# from src.beamprofiler import beam

# __all__ = ['analysis']
